      .486
      .model flat, stdcall
      option casemap :none

      include \masm32\include\windows.inc
      include \masm32\include\user32.inc
      include \masm32\include\kernel32.inc
      include \masm32\include\masm32.inc
      include \masm32\include\fpu.inc
      include \masm32\macros\macros.asm
      includelib \masm32\lib\masm32.lib
      includelib \masm32\lib\user32.lib
      includelib \masm32\lib\kernel32.lib
      includelib \masm32\lib\fpu.lib

      rand proto :DWORD
      intmean proto :DWORD
      intstd proto :DWORD
      
.data
      buffer      db 30 dup(0)
      randB       db 2048 dup(?)
      randD       dd 512 dup(?)
      fnB         db "randb.bin",0
      fnD         db "randd.bin",0
.code
start:
      ; Generate a 2048 byte stream from the rand
      ; procedure and write it to a file.
      xor   ebx,ebx
   @@:
      invoke rand,256      
      mov   [randB+ebx],al
      inc   ebx
      cmp   ebx,2048
      jb    @B      
      invoke write_disk_file,addr fnB,addr randB,2048

      ; Generate a 2048 byte stream directly from
      ; the counter and write it to a file.
      xor   ebx,ebx
   @@:
      call counter
      mov   [randD+ebx],edx
      mov   [randD+ebx+4],eax
      add   ebx,8
      cmp   ebx,2048
      jb    @B      
      invoke write_disk_file,addr fnD,addr randD,2048

      ; Calc and display the arithmetic mean for 100000
      ; random integers in the interval 0 to 99.
      invoke intmean,100
      invoke FpuFLtoA,0,5,addr buffer,SRC1_FPU or SRC2_DIMM
      print SADD("INT MEAN (49.5) = ")
      print addr buffer
      print SADD(13,10)

      ; Calc and display the standard deviation for 100000
      ; random integers in the interval 0 to 99.
      invoke intstd,100
      invoke FpuFLtoA,0,2,addr buffer,SRC1_FPU or SRC2_DIMM
      print SADD("INT STD (28.87) = ")
      print addr buffer
      print SADD(13,10)

      ; Calc and display the chi-square "statistic" for
      ; 100000 random integers in the interval 0 to 99.      
      call intchisq
      invoke FpuFLtoA,0,2,addr buffer,SRC1_FPU or SRC2_DIMM
      print SADD("INT CHI-SQ (80 to 120) = ")
      print addr buffer
      print SADD(13,10)
                 

      print SADD(13,10,"Esc to exit, other key for next",13,10,13,10)
   next:
      invoke rand,100
      invoke dw2a,eax,addr buffer

      print addr buffer
      print SADD(13,10)      
      call getch
      cmp   eax,27
      jne   next

      print SADD(13,10,"Press any normal key to exit...",13,10)
      call getch
      exit

; 
; This procedure implements a 64-bit, maximum-length LFSR
; counter with XNOR feedback. The basic concept and feedback
; tap selections were taken from XILINX application note 052,
; July 7,1996 (Version 1.1). The feedback taps are from SR
; stages 60,61,63,64 (EAX bits 4,3,1,0). Rather than performing
; an actual XNOR operation on the tap bits, the code sets the
; feedback bit based on the parity of the tap bits (set for
; even, clear for odd). The counter performs 64 shifts per
; call and returns the count in EDX:EAX.
; 
.data
      shiftreg dd 0,0
.code
counter proc
      push  ebx
      mov   eax,shiftreg
      mov   edx,shiftreg+4
      mov   ecx,64
   shiftLoop:
      xor   ebx,ebx
      test  eax,1ah
      jpo   @F
      mov   ebx,80000000h
   @@:      
      shr   edx,1
      rcr   eax,1
      or    edx,ebx
      dec   ecx
      jnz   shiftLoop
      mov   shiftreg,eax
      mov   shiftreg+4,edx      
      pop   ebx
      ret
counter endp

; 
; This procedure returns a random 32-bit integer in the
; interval 0 to base - 1.
;
; Thanks to pdixon for an easy way to avoid DIV :)
; 
rand proc base:DWORD
      call counter
      mov   ecx,base
      mul   ecx
      mov   eax,edx
      ret
rand endp

; 
; This procedure calculates the mean value for 100000 random
; 32-bit integers in the interval defined by the parameters
; and leaves the result on the FPU.
; 
intmean proc uses ebx base:DWORD
      
      LOCAL cnt:DWORD,sum:QWORD
      
      mov   cnt,100000
      mov   dword ptr[sum],0
      mov   dword ptr[sum+4],0      
      mov   ebx,cnt
   @@:
      invoke rand,base
      add   dword ptr[sum],eax
      adc   dword ptr[sum+4],0
      dec   ebx
      jnz   @B
      fild  sum
      fild  cnt
      fdiv
      ret
intmean endp

; 
; This procedure calculates the standard deviation for
; 100000 random 32-bit integers in the interval defined
; by the parameters and leaves the result on the FPU.
;
; * Reinitialize intsumsq before second call *
; 
.data
      intsumsq dt 0.0
.code
intstd proc uses ebx base:DWORD
      
      LOCAL cnt:DWORD,sum:QWORD,temp:DWORD
      
      mov   cnt,100000
      mov   dword ptr[sum],0
      mov   dword ptr[sum+4],0
      mov   ebx,cnt
   @@:
      invoke rand,base
      add   dword ptr[sum],eax
      adc   dword ptr[sum+4],0
      mov   temp,eax
      fild  temp
      fild  temp     ; ST=temp,ST(1)=temp      
      fmul           ; ST=temp*temp
      fld   intsumsq ; ST=intsumsq,ST(1)=temp*temp
      fadd           ; ST=intsumsq+temp*temp
      fstp  intsumsq ; intsumsq=intsumsq+temp*temp
      dec   ebx
      jnz   @B
      
      fild  sum
      fild  sum      ; ST=sum,ST(1)=sum
      fmul           ; ST=sum*sum
      fild  cnt      ; ST=cnt,ST(1)=sum*sum
      fdiv           ; ST=sum*sum/cnt
      fld   intsumsq ; ST=intsumsq,ST(1)=sum*sum/cnt
      fsubr          ; ST=intsumsq-sum*sum/cnt
      dec   cnt
      fild  cnt      ; ST=cnt-1,ST(1)=intsumsq-sum*sum/cnt
      fdiv           ; ST=(intsumsq-sum*sum/cnt)/(cnt-1)
      fsqrt          ; ST=SQR((ctrsumsq-sum*sum/cnt)/(cnt-1))
      ret
intstd endp

; 
; This procedure calculates the chi-square statistic for
; 10000 random positive integers in the interval 0-99
; and leaves the result on the FPU.
;
; Adapted (correctly I hope) from a Pascal example in
; Robert Sedgewick's Algorithms, Second Edition.
;
; Presumably, a good RNG will have a statistic that is
; within 2 * SQR(range) of the range (for N > 10r).
; 
intchisq proc uses ebx edi

      LOCAL cnt:DWORD,r:DWORD,sumsq:DWORD

      ; Allocate array to store frequencies.
      invoke Alloc,512
      or    eax,eax
      jnz   @F
      print SADD(13,10,"Alloc failed, skipping chi-square",13,10)
      ret
   @@:
      mov   edi,eax
      mov   cnt,10000
      mov   r,100
      mov   sumsq,0
      ; Init frequency array.
      xor   ebx,ebx
   @@:
      mov   dword ptr[edi+ebx*4],0
      inc   ebx
      cmp   ebx,r
      jb    @B
      ; Get frequencies.
      mov   ebx,cnt
   @@:      
      invoke rand,r
      inc   dword ptr[edi+eax*4]
      dec   ebx
      jnz   @B
      ; Get the sum of the squares of the frequencies.
      xor   ebx,ebx
   @@:
      mov   eax,[edi+ebx*4]
      mov   ecx,eax
      mul   ecx
      add   sumsq,eax
      inc   ebx
      cmp   ebx,r
      jb    @B
      ; Calc chi-square value.
      fild  r
      fild  sumsq    ; ST=sumsq,ST(1)=r
      fmul           ; ST=r*sumsq
      fild  cnt      ; ST=cnt,ST(1)=r*sumsq
      fdiv           ; ST=r*sumsq/cnt
      fild  cnt      ; ST=cnt,ST(1)=r*sumsq/cnt
      fsub           ; ST=(r*sumsq/cnt)-cnt
      ret
intchisq endp

; 
; This procedure waits for and reads a single
; character from stdin (no echo).
; 
getch proc
   
   LOCAL hInput :DWORD
   LOCAL char   :DWORD
   LOCAL bRead  :DWORD

   mov char,0
   invoke GetStdHandle,STD_INPUT_HANDLE
   mov   hInput,eax
   invoke FlushConsoleInputBuffer,hInput
   invoke SetConsoleMode,hInput,ENABLE_PROCESSED_INPUT
   invoke ReadConsole,hInput,ADDR char,1,ADDR bRead,NULL
   invoke SetConsoleMode,hInput,ENABLE_LINE_INPUT or \
                                ENABLE_ECHO_INPUT or \
                                ENABLE_PROCESSED_INPUT
   mov eax,char
   ret
getch endp
; 
end start